
#include "terrain.h"
#include "terrain_page.h"
#include "glincl.h"

#ifdef HAVE_SDL_SDL_IMAGE_H
#include <SDL/SDL_image.h>
#else
#include <SDL_image.h>
#endif

void Terrain::build()
{
	SDL_Surface *tmpData = IMG_Load( heightmapSource.c_str() );
 	if ( !tmpData ) {
		printf( "sorry, can't load!!!\n" );
		return;
	}		
	
	heightfield.height = tmpData->h;
	heightfield.width = tmpData->w;
	heightfield.pitch = tmpData->pitch;
	int dataSize = heightfield.pitch*heightfield.height;
	heightfield.data = new UByte[dataSize];
	memcpy( heightfield.data, tmpData->pixels, sizeof(UByte)*dataSize );
	SDL_FreeSurface( tmpData );
	//TODO: Check for modulo!
	pagesX = (heightfield.width/(pageSize-1) );
	pagesZ = (heightfield.height/(pageSize-1) );
	tex = new Texture( lightmapSource );
	
	printf( "creating terrain pages ( %d, %d )...", pagesX, pagesZ );
	pages = new pTerrainPage[pagesX*pagesZ];	
	for ( int x = 0; x < pagesX; ++x )
		for ( int z = 0; z < pagesZ; ++z )
			pages[z*pagesX+x] = createPage( x, z );
	printf( "looks good\n" );
	printf( "creating quad_tree data structure..." );
	root = createQuadTree( 0, 0, pagesX, pagesZ );
	activePages = NULL;
	printf( "looks good\n" );
}

pTerrainQuad Terrain::createQuadTree( int _x0, int _z0, int _x1, int _z1, int _depth )
{
	int _x01 = (_x1+_x0)/2; int _z01 = (_z1+_z0)/2;
	pTerrainQuad node;
	if ( _x1-_x0 == 1 ) {
		node = getPage( _x0, _z0 );
		node->setChildren( NULL, NULL, NULL, NULL );
	}
	else {
		assert( ( _x0 % 2 ) == 0 ); 
		assert( ( _x1 % 2 ) == 0 ); 
		assert( ( _z0 % 2 ) == 0 ); 
		assert( ( _z1 % 2 ) == 0 );		
		node = new TerrainQuad( this, _x0, _z0, _x1, _z1 );
		node->setChildren( 
			createQuadTree( _x0, _z0, _x01, _z01, _depth+1 ),
			createQuadTree( _x01, _z0, _x1, _z01, _depth+1 ), 
			createQuadTree( _x0, _z01, _x01, _z1, _depth+1 ), 
			createQuadTree( _x01, _z01, _x1, _z1, _depth+1 ) );
	}
	node->calculateBounds();
	return node;
}

pTerrainPage Terrain::createPage( int _xOffset, int _zOffset ) const
{
	pTerrainPage newPage = new TerrainPage( const_cast<Terrain*>( this ), _xOffset, _zOffset );
	newPage->setScale( scale );	
	newPage->setPosition( Triple( scale.x*_xOffset, 0.0f, scale.z*_zOffset ) );
	newPage->calculateErrors();
	
	//newPage->calculateMinMax();
	return newPage;
}

//TODO: The method could be optimized. We could get rid of the hidePages call
//		by setting a active flag in every pages that is visible. Then a walk
//		through the activePages list is sufficient to remove all invisible 
//		pages.
void Terrain::determineVisiblePages( pTerrainQuad _node )
{
	switch( _node->cull() ) {
		case Frustum::INTERSECT:
			//printf( "partially inside frustum\n" );
			if ( !_node->isChildless() ) {
				pTerrainQuad *children = _node->getChildren();
				for ( int i = 0; i < 4; ++i, ++children )
					determineVisiblePages( *children );
			}
			else {
				showPages( _node->getXOffset(), 
						   _node->getZOffset(), 1, 1 );
			}
			break;
		case Frustum::INSIDE:
			//printf( "fully inside frustum\n" );			
			showPages( _node->getXOffset(), 
					   _node->getZOffset(), 
					   _node->getWidth() , 
					   _node->getHeight() );
			break;
		case Frustum::OUTSIDE:
			cullCount+= (_node->getWidth()-1)*(_node->getHeight()-1);
			break;
	}
}

void Terrain::draw( )
{
	pTerrainPage page = NULL;
	
	frustum->extractPlanes();
	
	glEnableClientState( GL_VERTEX_ARRAY );
	glEnableClientState( GL_TEXTURE_COORD_ARRAY ); 
	glEnable( GL_TEXTURE_2D );
	glEnable( GL_DEPTH_TEST );
	glEnable( GL_CULL_FACE );

	
	glBindTexture( GL_TEXTURE_2D, tex->getTexture() );
	glDisable( GL_LIGHTING );

	glEnable( GL_COLOR_MATERIAL );
	glColor3f( 1.0f, 1.0f, 1.0f );
	cullCount = 0;
	
	// Lets see which pages are visible.
	determineVisiblePages( root );
		
	int count = 0;
	page = activePages;	
	
	while ( page ) {
		if ( !page->isActive() ) {
			pTerrainPage tmp = page;
			page = tmp->getNext();
			tmp->deactivate();
			continue;
		}	
		count++;
		page->updateTesselation();
		//Draw the page. TODO: It would be nice if all the pages with an LOD of 4 could be merged
		//in one index buffer. This would give us a speed boost for sure...
		page->draw();
		page = page->getNext();
	}
	
	//printf( "%d pages activated, %d pages deactivated\n", activatedCount, deactivatedCount );
	activatedCount = 0; deactivatedCount = 0;
	//float percentage = (float)cullCount/(float)(pagesX*pagesZ)*100.0f;
	//printf( "culled %f%% terrain pages away\n",  percentage );
}

void Terrain::getAltitude( Triple& _alt, Triple& _normal )
{
	
}

void Terrain::showPages( int _x0, int _z0, int _width, int _height )
{
	for ( int x = 0; x < _width; ++x ) {
		for ( int z = 0; z < _height; ++z ) {
			pTerrainPage page = getPage( _x0+x, _z0+z );
			page->chooseLOD( );
			page->show( );
		}
	}
}
