/*
	orxonox - the future of 3D-vertical-scrollers
 
	Copyright (C) 2006 orx
 
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2, or (at your option)
	any later version.
 
	### File Specific:
	main programmer: Marco Biasini
 
 */


#ifndef TERRAIN_PAGE_H
#define TERRAIN_PAGE_H

#include "terrain_quad.h"
#include <stdio.h>


class TerrainPage;
class Terrain;

typedef TerrainPage *pTerrainPage;

typedef struct {
	Triple	correct;
	Triple	real;
	float	diff;
} LODError, *pLODError;

class TerrainPage : public TerrainQuad {
	public:
		const static int MAX_LODS = 5;
		/**
		 * Creates a new terrain page with its lower left corner set
		 * to C = ( _xOffset, _zOffset ), where the two values specify
		 * the offset in the height-map.
		 * The size of the page, as well as the scaling factors are read
		 * from the _owner terrain page.
		 */
		TerrainPage( Terrain *_owner, int _xOffset, int _zOffset );
		
		inline bool isActive() { return active; }
		~TerrainPage( )
		{ 
			if ( isVisible )
				hide(); 
		}
		
		/**
		 * Makes the terrain look as if it were created with the given level of 
		 * detail.
		 */
		void mimick( int _level ) {}
		
		void drawBox();
		virtual void calculateBounds();		
		/**
		 * Sets the visibility to _flag. If the visibility changed, the vertex and index
		 * arrays are allocated or freed, respectively.
		 */
		inline void setVisibility( bool _flag );

		/**
		 * If
		 */
		void show( );
		void hide( ); 
		/**
		 * 
		 */
		void updateTesselation( );
		
		/**
		 * Returns the current tesselation level.
		 */
		int getLOD() { return currentLOD; }
		
		/**
		 * Returns the curren tween factor. This is a number between 0.0f and 1.0f
		 */
		float getTween() { return 0.0f; }
		
		/**
		 * Determines the new LOD which should be used by this terrain page based on
		 * the distance from the camera. 
		 * Note: No geometry is updated in this method. You need to call 
		 * updateTesselation() in order to see a change in geometry. This method is
		 * just a recommandation.
		 */
		int chooseLOD();
		/**
		 * If the terrain pages tesselation level changed between the last and the
		 * current frame, this function returns true, else you'll get false as the
		 * return value.
		 */
		bool isDirty() { return true; }
		
		/**
		 * Calculates the maximal errors for every LOD.
		 */
		void calculateErrors();
		
		/**
		 * Calculates the error for the given LOD. We just need to know the "worst"
		 * vertex for choosing an appropriate LOD.
		 */
		void calculateError( int _lod );
		/**
		 * Tests if the terrain page would cull against the viewing frustum.
		 */
		bool cull( );
		
		void calculateMinMax();
		
		/**
		 * Sets the neighbors of this terrain page. pass null if a neighbor if this
		 * pages is at the border.
		 */
		inline void setNeighbors( pTerrainPage _left, pTerrainPage _right, 
			pTerrainPage _top, pTerrainPage _bottom )
		{
			left = _left; right = _right; top = _top; bottom = _bottom;
		}
		inline void setPosition( const Triple& _pos )
		{
			position.x = _pos.x;
			position.y = _pos.y;
			position.z = _pos.z;
		}
		
		/**
		 *  Does what exactly what the name says and nothing more.
		 */
		void draw( );
		
		inline pTerrainPage getNext() { return next; }
		inline pTerrainPage getPrevious() { return previous; }			
		inline int getWantedLOD() { return wantedLOD; }
		void deactivate();
		void activate();		
		protected:
			
			inline void addIndex( unsigned short _index );
	
			inline void addAgain();
			
			void getCoord( int _x, int _z, TexCoord& _coord) const;
			/** 
			 * Fills _vertex with the vertex information at index.
			 */
			void getVertex( int _x, int _z, Triple& _vertex ) const;
			
			/**
			* Use this method to safely get a vertex at location ( _x, _z ). If it wasn't
			* created before, this method does that for you. 
			*/			
			short getIndex( int _x, int _z );
			
		   /**
			* Generates the tesselation for the given level of detail.
			*/
			void tesselate( int _lod );			
			
			float getAltitude( int _x, int _z ) const;
			
			int							currentLOD,
										wantedLOD;
			pTerrainPage 				left, 
										right,
										top,
										bottom;
			bool						forceTesselation;
			bool						active;
			Triple						*vertices;
			unsigned short				*indices;
			unsigned short				*indexHash;
			TexCoord					*coords;
			int							numIndices;
			int							numVertices;
			bool						isVisible;
			pTerrainPage				next;
			pTerrainPage				previous;
			LODError					*errors;
			Triple						position;
};

inline void TerrainPage::setVisibility( bool _flag )
{
	if ( _flag ) {
		if ( !isVisible ) {
			isVisible = true;
			show( );
		}
	}
	else {
		if ( isVisible ) {
			isVisible = false;
			hide( );
		}
	}
}

inline void TerrainPage::addIndex( unsigned short _index )
{
	indices[numIndices] = _index; numIndices++;
}

inline void TerrainPage::addAgain() 
{ 
	indices[numIndices] = indices[numIndices-1]; numIndices++;				
}

#endif